class Entity:
    """Entity

    Parent class for features that are in the 'entity' family.
    """

    def total_number_of_named_entities(
        SE: object
        ) -> int:
        """
        returns the number of named entities
        """
        try:
            return SE.total_number_of_named_entities_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                doc_total += 1
            SE.total_number_of_named_entities_ = doc_total
            return SE.total_number_of_named_entities_
   
    def total_number_of_named_entities_person(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are PERSON (People, including fictional)
        """
        try:
            return SE.total_number_of_named_entities_person_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "PERSON":
                    doc_total += 1
            SE.total_number_of_named_entities_person_ = doc_total
            return SE.total_number_of_named_entities_person_
        
    def total_number_of_named_entities_norp(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are NORP (Nationalities or religious or political groups)
        """
        try:
            return SE.total_number_of_named_entities_norp_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "NORP":
                    doc_total += 1
            SE.total_number_of_named_entities_norp_ = doc_total
            return SE.total_number_of_named_entities_norp_

    def total_number_of_named_entities_fac(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are FAC (Buildings, airports, highways, bridges, etc.)
        """
        try:
            return SE.total_number_of_named_entities_fac_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "FAC":
                    doc_total += 1
            SE.total_number_of_named_entities_fac_ = doc_total
            return SE.total_number_of_named_entities_fac_

    def total_number_of_named_entities_org(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are ORG (Companies, agencies, institutions, etc.)
        """
        try:
            return SE.total_number_of_named_entities_org_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "ORG":
                    doc_total += 1
            SE.total_number_of_named_entities_org_ = doc_total
            return SE.total_number_of_named_entities_org_
    
    def total_number_of_named_entities_gpe(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are GPE (Countries, cities, states.)
        """
        try:
            return SE.total_number_of_named_entities_gpe_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "GPE":
                    doc_total += 1
            SE.total_number_of_named_entities_gpe_ = doc_total
            return SE.total_number_of_named_entities_gpe_
    
    def total_number_of_named_entities_loc(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are LOC (Non-GPE locations, mountain ranges, bodies of water.)
        """
        try:
            return SE.total_number_of_named_entities_loc_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "LOC":
                    doc_total += 1
            SE.total_number_of_named_entities_loc_ = doc_total
            return SE.total_number_of_named_entities_loc_
    
    def total_number_of_named_entities_product(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are PRODUCT (Objects, vehicles, foods, etc. (Not services.)
        """
        try:
            return SE.total_number_of_named_entities_product_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "PRODUCT":
                    doc_total += 1
            SE.total_number_of_named_entities_product_ = doc_total
            return SE.total_number_of_named_entities_product_
    
    def total_number_of_named_entities_event(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are EVENT (Named hurricanes, battles, wars, sports events, etc.)
        """
        try:
            return SE.total_number_of_named_entities_event_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "EVENT":
                    doc_total += 1
            SE.total_number_of_named_entities_event_ = doc_total
            return SE.total_number_of_named_entities_event_
    
    def total_number_of_named_entities_art(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are WORK_OF_ART (Titles of books, songs, etc.)
        """
        try:
            return SE.total_number_of_named_entities_art_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "WORK_OF_ART":
                    doc_total += 1
            SE.total_number_of_named_entities_art_ = doc_total
            return SE.total_number_of_named_entities_art_
    
    def total_number_of_named_entities_law(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are LAW (Named documents made into laws.)
        """
        try:
            return SE.total_number_of_named_entities_law_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "LAW":
                    doc_total += 1
            SE.total_number_of_named_entities_law_ = doc_total
            return SE.total_number_of_named_entities_law_
    
    def total_number_of_named_entities_language(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are LANGUAGE (Any named language.)
        """
        try:
            return SE.total_number_of_named_entities_language_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "LANGUAGE":
                    doc_total += 1
            SE.total_number_of_named_entities_language_ = doc_total
            return SE.total_number_of_named_entities_language_
    
    def total_number_of_named_entities_date(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are DATE (Absolute or relative dates or periods.)
        """
        try:
            return SE.total_number_of_named_entities_date_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "DATE":
                    doc_total += 1
            SE.total_number_of_named_entities_date_ = doc_total
            return SE.total_number_of_named_entities_date_
        
    def total_number_of_named_entities_time(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are TIME (Times smaller than a day.)
        """
        try:
            return SE.total_number_of_named_entities_time_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "TIME":
                    doc_total += 1
            SE.total_number_of_named_entities_time_ = doc_total
            return SE.total_number_of_named_entities_time_
    
    def total_number_of_named_entities_percent(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are PERCENT (Percentage, including ”%“.)
        """
        try:
            return SE.total_number_of_named_entities_percent_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "PERCENT":
                    doc_total += 1
            SE.total_number_of_named_entities_percent_ = doc_total
            return SE.total_number_of_named_entities_percent_
    
    def total_number_of_named_entities_money(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are MONEY (Monetary values, including unit.)
        """
        try:
            return SE.total_number_of_named_entities_money_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "PERCENT":
                    doc_total += 1
            SE.total_number_of_named_entities_money_ = doc_total
            return SE.total_number_of_named_entities_money_
    
    def total_number_of_named_entities_quantity(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are QUANTITY (Measurements, as of weight or distance.)
        """
        try:
            return SE.total_number_of_named_entities_quantity_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "QUANTITY":
                    doc_total += 1
            SE.total_number_of_named_entities_quantity_ = doc_total
            return SE.total_number_of_named_entities_quantity_
    
    def total_number_of_named_entities_ordinal(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are ORDINAL (“first”, “second”, etc.)
        """
        try:
            return SE.total_number_of_named_entities_ordinal_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "ORDINAL":
                    doc_total += 1
            SE.total_number_of_named_entities_ordinal_ = doc_total
            return SE.total_number_of_named_entities_ordinal_
    
    def total_number_of_named_entities_cardinal(
        SE: object
        ) -> int:
        """
        returns the number of named entities that are CARDINAL (Numerals that do not fall under another type.)
        """
        try:
            return SE.total_number_of_named_entities_cardinal_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                if ent.label_ == "CARDINAL":
                    doc_total += 1
            SE.total_number_of_named_entities_cardinal_ = doc_total
            return SE.total_number_of_named_entities_cardinal_


# -*- coding: UTF-8 -*-
import os
from typing import Union, Dict, List
import sys

import spacy
import pandas as pd

from lftk.utils import get_feature_map, get_pandas_row, convert_ndjson_to_pd
from lftk.derivation_collector import DerivationCollector


CURRENT_PATH = os.path.abspath(os.path.dirname(__file__))
FEATURE_MAP_PATH = os.path.join(CURRENT_PATH, 'resources/feature_map.json')
AOA_KUP_PATH = os.path.join(CURRENT_PATH, 'resources/AoA_ratings_kup.csv')
AOA_BRY_PATH = os.path.join(CURRENT_PATH, 'resources/AoA_ratings_bry.csv')
SUBTLEX_US_PATH = os.path.join(CURRENT_PATH, 'resources/subtlex_us.csv')


class Extractor:
    """Main object

    Saves your spaCy doc objects to extract linguistic features from. As you declare this object, pass in spaCy doc(s).
        
    Parameters
    ----------
    docs : Union[spacy.tokens.doc.Doc, List[spacy.tokens.doc.Doc]]
        spaCy doc object or a list of spacy doc objects

    Examples
    ----------
    >>> import spacy
    >>> import lftk
    >>> 
    >>> nlp = spacy.load("en_core_web_sm")
    >>> doc1 = nlp("I think effects computers have on people are great!")
    >>> doc2 = nlp("I like drinking coffee...")
    >>> 
    >>> LFTK = lftk.Extractor(docs = [doc1, doc2])


    """
    def __init__(
        self, 
        docs: Union[spacy.tokens.doc.Doc, List[spacy.tokens.doc.Doc]]
        ) -> None:
        # Type adjustment
        if type(docs) is not list: docs = [docs]
        self.docs = docs
        # dictionary of dictionary that contains "key" as the parent key and the others ("name", "type") as children keys
        self.feature_map = get_feature_map(path = FEATURE_MAP_PATH)
        self.customize()

    def customize(
        self,
        stop_words: bool = True,
        punctuations: bool = True,
        round_decimal: int = 3,
        ) -> None:
        """Global customization
    
        Customizes all LFTK functions to extract features based on these options. This exclude some special functions that are intentionally designed to override global options.
        
        Parameters
        ----------
        stop_words : bool (default = True)
            Selection whether to include stop words for feature extraction
        punctuations : bool (default = True)
            Selection whether to include punctuations for feature extraction
        round_decimal : int (default = 3)
            The max number of decimal digits to return (for extracted feature values)
    
        Examples
        ----------
        >>> import spacy
        >>> import lftk
        >>> 
        >>> nlp = spacy.load("en_core_web_sm")
        >>> doc = nlp("I think effects computers have on people are great!")
        >>> 
        >>> LFTK = lftk.Extractor(docs = doc)
        >>> 
        >>> LFTK.customize(stop_words=True, punctuations=False, round_decimal=3)
        >>> output = LFTK.extract(features = ["a_word_ps", "a_kup_pw", "n_noun"])
        >>> # {'a_word_ps': 9.0, 'a_kup_pw': 5.323, 'n_noun': 3}
        >>> 
        >>> LFTK.customize(stop_words=False, punctuations=False, round_decimal=2)
        >>> output = LFTK.extract(features = ["a_word_ps", "a_kup_pw", "n_noun"])
        >>> # {'a_word_ps': 5.0, 'a_kup_pw': 9.6, 'n_noun': 3}


        """
        self.options = {
            'stop_words': stop_words,
            'punctuations': punctuations,
            'round_decimal': round_decimal
        }

    def extract(
        self, 
        features: Union[str, list] = "*"
        ) -> Union[
            Dict[str, Union[float,int]], 
            List[Dict[str, Union[float,int]]]
            ]:
        """Extract function for select features
    
        Extracts feature(s) that are passed in by the user. All spaCy docs that were saved during LFTK Extractor declaration are used. This function extracts all features available in LFTK by default. 
        
        This function only accepts feature key(s) (e.g., "t_word", "a_kup_pw").
        
        Parameters
        ----------
        features : Union[str, list] (default = "*")
            A single feature key or a list of feature keys. Passing "*" extracts all available features.

        Returns
        ----------
        result : Union[Dict[str, Union[float,int]], List[Dict[str, Union[float,int]]]]
            A dictionary or a list of dictionaries, depending on the number of spaCy docs you passed in during the creation of LFTK Extractor object.

        Examples
        ----------
        >>> import spacy
        >>> import lftk
        >>> 
        >>> nlp = spacy.load("en_core_web_sm")
        >>> doc1 = nlp("I think effects computers have on people are great!")
        >>> doc2 = nlp("I like drinking coffee...")
        >>> 
        >>> LFTK = lftk.Extractor(docs = doc1)
        >>> 
        >>> output = LFTK.extract(features = ["a_word_ps", "a_kup_pw", "n_noun"])
        >>> # {'a_word_ps': 10.0, 'a_kup_pw': 4.791, 'n_noun': 3}
        >>> 
        >>> LFTK = lftk.Extractor(docs = [doc1, doc2])
        >>> 
        >>> output = LFTK.extract(features = ["a_word_ps", "a_kup_pw", "n_noun"])
        >>> # [{'a_word_ps': 10.0, 'a_kup_pw': 4.791, 'n_noun': 3}, {'a_word_ps': 5.0, 'a_kup_pw': 2.284, 'n_noun': 2}]
        """
        # Type adjustment
        if type(features) is not list: features = [features]
        # Iterate user input
        all_results = []
        for doc in self.docs:
            extractor = SingleExtractor(doc, self.feature_map, self.options)
            result = extractor.run(features)
            all_results.append(result)
        # Type adjustment
        if len(all_results) == 1:
            all_results = all_results[0]
        return all_results


class SingleExtractor:
    def __init__(
        self, 
        doc: spacy.tokens.doc.Doc,
        feature_map: Dict[str, Dict[str, str]],
        options: Dict[str, Union[str, bool]]
        ) -> None:
        """
            Initialize pipeline
            input :
            - self
            - doc: single text inside spacy doc object
            saves :
            - self.doc: spacy doc object
        """
        self.doc = doc
        self.feature_map = feature_map
        self.options = options
        self.paths = {
            'FEATURE_MAP_PATH' : FEATURE_MAP_PATH,
            'AOA_KUP_PATH' : AOA_KUP_PATH,
            'AOA_BRY_PATH' : AOA_BRY_PATH,
            'SUBTLEX_US_PATH' : SUBTLEX_US_PATH,
        }

    def run(
        self, 
        features: list,
        ) -> Dict[str, Union[float,int]]:
        """
            Calculate linguistic feature(s) from spaCy doc
            input :
            - self
            - features: features to be extracted
            returns :
            - result: extracted linguistic feature(s)
        """
        # Iterate through given features
        result = {}
        if features != ["*"]:
            for feature_key in features:
                # Get feature name from feature key
                feature_name = self.feature_map[feature_key]['name']
                # Get feature function from feature name
                feature_fn = getattr(DerivationCollector, feature_name)
                # Run feature function in DerivationCollector object
                result[f'{feature_key}'] = round(feature_fn(self), self.options['round_decimal'])
        else:
            # Extract all features
            for feature_key in self.feature_map:
                # Get feature name from feature key
                feature_name = self.feature_map[feature_key]['name']
                # Get feature function from feature name
                feature_fn = getattr(DerivationCollector, feature_name)
                # Run feature function in DerivationCollector object
                result[f'{feature_key}'] = round(feature_fn(self), self.options['round_decimal'])
        return result

def search_features(
    domain: str = "*",
    family: str = "*",
    language: str ="*",
    return_format: str = "list_dict"
    ) -> Union[List[dict], pd.core.frame.DataFrame]:
    """Search features
    
    Returns available linguistic features that match user-specified attributes. Putting "*" on any attribute is analogous to "any". You can use this function to produce list of feature keys and pass them into LFTK Extractor's extract() function (see above).
    
    Parameters
    ----------
    domain : str (default = "*")
        A single domain name (e.g., "surface", "lexico-semantics")
    family : str (default = "*")
        A single family name (e.g., "worddiff", "avgwordsent")
    language : str (default = "*")
        A single supported language (e.g., "worddiff", "avgwordsent")
    return_format : str (default = "list_dict")
        Select how the searched features should be returned. The available options are "list_dict", "pandas", or "list_key".
    
    Returns
    ----------
    result : Union[List[dict], pd.core.frame.DataFrame]
        Available features searched with user-given conditions

    Examples
    ----------
    >>> import lftk
    >>> 
    >>> output = lftk.search_features(domain = 'surface', family = "avgwordsent", language="general", return_format = "list_dict")
    >>> # [{'key': 'a_word_ps', 'name': 'average_number_of_words_per_sentence', 'formulation': 'derivation', 'domain': 'surface', 'family': 'avgwordsent', 'language': 'general'}, {'key': 'a_char_ps', 'name': 'average_number_of_characters_per_sentence', 'formulation': 'derivation', 'domain': 'surface', 'family': 'avgwordsent', 'language': 'general'}, {'key': 'a_char_pw', 'name': 'average_number_of_characters_per_word', 'formulation': 'derivation', 'domain': 'surface', 'family': 'avgwordsent', 'language': 'general'}]
    >>> 
    >>> output = lftk.search_features(domain = 'surface', family = "avgwordsent", language="general", return_format = "list_dict")
    >>> #           key                                       name formulation   domain       family language
    >>>  #9   a_word_ps       average_number_of_words_per_sentence  derivation  surface  avgwordsent  general
    >>> # 10  a_char_ps  average_number_of_characters_per_sentence  derivation  surface  avgwordsent  general
    >>> # 11  a_char_pw      average_number_of_characters_per_word  derivation  surface  avgwordsent  general
    >>> 
    >>> output = lftk.search_features(domain = 'surface', family = "avgwordsent", language="general", return_format = "list_dict")
    >>> # ['a_word_ps', 'a_char_ps', 'a_char_pw']
    """
    feature_df = convert_ndjson_to_pd(path = FEATURE_MAP_PATH)
    if domain != "*":
        feature_df = get_pandas_row(feature_df, "domain", domain, True)
    if family != "*":
        feature_df = get_pandas_row(feature_df, "family", family, True)
    if language != "*":
        feature_df = get_pandas_row(feature_df, "language", language, True)

    if return_format == "list_dict":
        result = feature_df.to_dict('records')
    elif return_format == "pandas":
        result = feature_df
    elif return_format == "list_key":
        list_dict = feature_df.to_dict('records')
        result = [dict_['key'] for dict_ in list_dict]
    else:
        raise ValueError('return_format must be "list_dict", "pandas", or "list_key"')

    return result


from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class AvgEntity(FoundationCollector):
    """AvgEntity

    Parent class for features that are in the 'wordsent' family.
    """

    def average_number_of_named_entities_per_word(
        SE: object,
        ) -> float:
        """
        returns the value of (total number of named entities / total number of words)
        """
        try:
            return SE.average_number_of_named_entities_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_ = \
                FoundationCollector.total_number_of_named_entities(SE)
            SE.average_number_of_named_entities_per_word_ = \
                safe_division(
                    total_number_of_named_entities_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_per_word_
    
    def average_number_of_named_entities_person_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are PERSON / total number of words) -> People, including fictional
        """
        try:
            return SE.average_number_of_named_entities_person_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_person_ = \
                FoundationCollector.total_number_of_named_entities_person(SE)
            SE.average_number_of_named_entities_person_per_word_ = \
                safe_division(
                    total_number_of_named_entities_person_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_person_per_word_
        
    def average_number_of_named_entities_norp_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are NORP / total number of words) -> Nationalities or religious or political groups
        """
        try:
            return SE.average_number_of_named_entities_norp_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_norp_ = \
                FoundationCollector.total_number_of_named_entities_norp(SE)
            SE.average_number_of_named_entities_norp_per_word_ = \
                safe_division(
                    total_number_of_named_entities_norp_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_norp_per_word_

    def average_number_of_named_entities_fac_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are FAC / total number of words) -> Buildings, airports, highways, bridges, etc.
        """
        try:
            return SE.average_number_of_named_entities_fac_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_fac_ = \
                FoundationCollector.total_number_of_named_entities_fac(SE)
            SE.average_number_of_named_entities_fac_per_word_ = \
                safe_division(
                    total_number_of_named_entities_fac_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_fac_per_word_

    def average_number_of_named_entities_org_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are ORG / total number of words) -> Companies, agencies, institutions, etc.
        """
        try:
            return SE.average_number_of_named_entities_org_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_org_ = \
                FoundationCollector.total_number_of_named_entities_org(SE)
            SE.average_number_of_named_entities_org_per_word_ = \
                safe_division(
                    total_number_of_named_entities_org_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_org_per_word_
    
    def average_number_of_named_entities_gpe_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are GPE / total number of words) -> Countries, cities, states.
        """
        try:
            return SE.average_number_of_named_entities_gpe_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_gpe_ = \
                FoundationCollector.total_number_of_named_entities_gpe(SE)
            SE.average_number_of_named_entities_gpe_per_word_ = \
                safe_division(
                    total_number_of_named_entities_gpe_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_gpe_per_word_
    
    def average_number_of_named_entities_loc_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are LOC / total number of words) -> Non-GPE locations, mountain ranges, bodies of water.
        """
        try:
            return SE.average_number_of_named_entities_loc_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_loc_ = \
                FoundationCollector.total_number_of_named_entities_loc(SE)
            SE.average_number_of_named_entities_loc_per_word_ = \
                safe_division(
                    total_number_of_named_entities_loc_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_loc_per_word_
    
    def average_number_of_named_entities_product_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are PRODUCT / total number of words) -> Objects, vehicles, foods, etc., but not services.
        """
        try:
            return SE.average_number_of_named_entities_product_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_product_ = \
                FoundationCollector.total_number_of_named_entities_product(SE)
            SE.average_number_of_named_entities_product_per_word_ = \
                safe_division(
                    total_number_of_named_entities_product_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_product_per_word_
    
    def average_number_of_named_entities_event_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are EVENT / total number of words) -> Named hurricanes, battles, wars, sports events, etc.
        """
        try:
            return SE.average_number_of_named_entities_event_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_event_ = \
                FoundationCollector.total_number_of_named_entities_event(SE)
            SE.average_number_of_named_entities_event_per_word_ = \
                safe_division(
                    total_number_of_named_entities_event_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_event_per_word_
    
    def average_number_of_named_entities_art_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are WORK_OF_ART / total number of words) -> Titles of books, songs, etc.
        """
        try:
            return SE.average_number_of_named_entities_art_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_art_ = \
                FoundationCollector.total_number_of_named_entities_art(SE)
            SE.average_number_of_named_entities_art_per_word_ = \
                safe_division(
                    total_number_of_named_entities_art_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_art_per_word_
    
    def average_number_of_named_entities_law_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are LAW / total number of words) -> Named documents made into laws.
        """
        try:
            return SE.average_number_of_named_entities_law_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_law_ = \
                FoundationCollector.total_number_of_named_entities_law(SE)
            SE.average_number_of_named_entities_law_per_word_ = \
                safe_division(
                    total_number_of_named_entities_law_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_law_per_word_
    
    def average_number_of_named_entities_language_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are LANGUAGE / total number of words) -> Any named language.
        """
        try:
            return SE.average_number_of_named_entities_language_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_language_ = \
                FoundationCollector.total_number_of_named_entities_language(SE)
            SE.average_number_of_named_entities_language_per_word_ = \
                safe_division(
                    total_number_of_named_entities_language_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_language_per_word_
    
    def average_number_of_named_entities_date_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are DATE / total number of words) -> Absolute or relative dates or periods.
        """
        try:
            return SE.average_number_of_named_entities_date_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_date_ = \
                FoundationCollector.total_number_of_named_entities_date(SE)
            SE.average_number_of_named_entities_date_per_word_ = \
                safe_division(
                    total_number_of_named_entities_date_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_date_per_word_
        
    def average_number_of_named_entities_time_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are TIME / total number of words) -> Times smaller than a day.
        """
        try:
            return SE.average_number_of_named_entities_time_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_time_ = \
                FoundationCollector.total_number_of_named_entities_time(SE)
            SE.average_number_of_named_entities_time_per_word_ = \
                safe_division(
                    total_number_of_named_entities_time_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_time_per_word_
    
    def average_number_of_named_entities_percent_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are PERCENT / total number of words) -> Percentage, including ”%“.
        """
        try:
            return SE.average_number_of_named_entities_percent_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_percent_ = \
                FoundationCollector.total_number_of_named_entities_percent(SE)
            SE.average_number_of_named_entities_percent_per_word_ = \
                safe_division(
                    total_number_of_named_entities_percent_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_percent_per_word_
    
    def average_number_of_named_entities_money_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are MONEY / total number of words) -> Monetary values, including unit.
        """
        try:
            return SE.average_number_of_named_entities_money_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_money_ = \
                FoundationCollector.total_number_of_named_entities_money(SE)
            SE.average_number_of_named_entities_money_per_word_ = \
                safe_division(
                    total_number_of_named_entities_money_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_money_per_word_
    
    def average_number_of_named_entities_quantity_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are QUANTITY / total number of words) -> Measurements, as of weight or distance.
        """
        try:
            return SE.average_number_of_named_entities_quantity_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_quantity_ = \
                FoundationCollector.total_number_of_named_entities_quantity(SE)
            SE.average_number_of_named_entities_quantity_per_word_ = \
                safe_division(
                    total_number_of_named_entities_quantity_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_quantity_per_word_
    
    def average_number_of_named_entities_ordinal_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are ORDINAL / total number of words) -> “first”, “second”, etc.
        """
        try:
            return SE.average_number_of_named_entities_ordinal_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_ordinal_ = \
                FoundationCollector.total_number_of_named_entities_ordinal(SE)
            SE.average_number_of_named_entities_ordinal_per_word_ = \
                safe_division(
                    total_number_of_named_entities_ordinal_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_ordinal_per_word_
    
    def average_number_of_named_entities_cardinal_per_word(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are CARDINAL / total number of words) -> Numerals that do not fall under another type.
        """
        try:
            return SE.average_number_of_named_entities_cardinal_per_word_ 
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_cardinal_ = \
                FoundationCollector.total_number_of_named_entities_cardinal(SE)
            SE.average_number_of_named_entities_cardinal_per_word_ = \
                safe_division(
                    total_number_of_named_entities_cardinal_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_cardinal_per_word_
        
    def average_number_of_named_entities_per_sentence(
        SE: object,
        ) -> float:
        """
        returns the value of (total number of named entities / total number of sentences)
        """
        try:
            return SE.average_number_of_named_entities_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_ = \
                FoundationCollector.total_number_of_named_entities(SE)
            SE.average_number_of_named_entities_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_per_sentence_
    
    def average_number_of_named_entities_person_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are PERSON / total number of sentences) -> People, including fictional
        """
        try:
            return SE.average_number_of_named_entities_person_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_person_ = \
                FoundationCollector.total_number_of_named_entities_person(SE)
            SE.average_number_of_named_entities_person_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_person_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_person_per_sentence_
        
    def average_number_of_named_entities_norp_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are NORP / total number of sentences) -> Nationalities or religious or political groups
        """
        try:
            return SE.average_number_of_named_entities_norp_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_norp_ = \
                FoundationCollector.total_number_of_named_entities_norp(SE)
            SE.average_number_of_named_entities_norp_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_norp_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_norp_per_sentence_

    def average_number_of_named_entities_fac_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are FAC / total number of sentences) -> Buildings, airports, highways, bridges, etc.
        """
        try:
            return SE.average_number_of_named_entities_fac_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_fac_ = \
                FoundationCollector.total_number_of_named_entities_fac(SE)
            SE.average_number_of_named_entities_fac_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_fac_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_fac_per_sentence_

    def average_number_of_named_entities_org_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are ORG / total number of sentences) -> Companies, agencies, institutions, etc.
        """
        try:
            return SE.average_number_of_named_entities_org_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_org_ = \
                FoundationCollector.total_number_of_named_entities_org(SE)
            SE.average_number_of_named_entities_org_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_org_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_org_per_sentence_
    
    def average_number_of_named_entities_gpe_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are GPE / total number of sentences) -> Countries, cities, states.
        """
        try:
            return SE.average_number_of_named_entities_gpe_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_gpe_ = \
                FoundationCollector.total_number_of_named_entities_gpe(SE)
            SE.average_number_of_named_entities_gpe_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_gpe_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_gpe_per_sentence_
    
    def average_number_of_named_entities_loc_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are LOC / total number of sentences) -> Non-GPE locations, mountain ranges, bodies of water.
        """
        try:
            return SE.average_number_of_named_entities_loc_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_loc_ = \
                FoundationCollector.total_number_of_named_entities_loc(SE)
            SE.average_number_of_named_entities_loc_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_loc_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_loc_per_sentence_
    
    def average_number_of_named_entities_product_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are PRODUCT / total number of sentences) -> Objects, vehicles, foods, etc., but not services.
        """
        try:
            return SE.average_number_of_named_entities_product_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_product_ = \
                FoundationCollector.total_number_of_named_entities_product(SE)
            SE.average_number_of_named_entities_product_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_product_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_product_per_sentence_
    
    def average_number_of_named_entities_event_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are EVENT / total number of sentences) -> Named hurricanes, battles, wars, sports events, etc.
        """
        try:
            return SE.average_number_of_named_entities_event_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_event_ = \
                FoundationCollector.total_number_of_named_entities_event(SE)
            SE.average_number_of_named_entities_event_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_event_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_event_per_sentence_
    
    def average_number_of_named_entities_art_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are WORK_OF_ART / total number of sentences) -> Titles of books, songs, etc.
        """
        try:
            return SE.average_number_of_named_entities_art_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_art_ = \
                FoundationCollector.total_number_of_named_entities_art(SE)
            SE.average_number_of_named_entities_art_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_art_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_art_per_sentence_
    
    def average_number_of_named_entities_law_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are LAW / total number of sentences) -> Named documents made into laws.
        """
        try:
            return SE.average_number_of_named_entities_law_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_law_ = \
                FoundationCollector.total_number_of_named_entities_law(SE)
            SE.average_number_of_named_entities_law_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_law_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_law_per_sentence_
    
    def average_number_of_named_entities_language_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are LANGUAGE / total number of sentences) -> Any named language.
        """
        try:
            return SE.average_number_of_named_entities_language_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_language_ = \
                FoundationCollector.total_number_of_named_entities_language(SE)
            SE.average_number_of_named_entities_language_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_language_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_language_per_sentence_
    
    def average_number_of_named_entities_date_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are DATE / total number of sentences) -> Absolute or relative dates or periods.
        """
        try:
            return SE.average_number_of_named_entities_date_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_date_ = \
                FoundationCollector.total_number_of_named_entities_date(SE)
            SE.average_number_of_named_entities_date_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_date_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_date_per_sentence_
        
    def average_number_of_named_entities_time_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are TIME / total number of sentences) -> Times smaller than a day.
        """
        try:
            return SE.average_number_of_named_entities_time_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_time_ = \
                FoundationCollector.total_number_of_named_entities_time(SE)
            SE.average_number_of_named_entities_time_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_time_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_time_per_sentence_
    
    def average_number_of_named_entities_percent_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are PERCENT / total number of sentences) -> Percentage, including ”%“.
        """
        try:
            return SE.average_number_of_named_entities_percent_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_percent_ = \
                FoundationCollector.total_number_of_named_entities_percent(SE)
            SE.average_number_of_named_entities_percent_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_percent_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_percent_per_sentence_
    
    def average_number_of_named_entities_money_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are MONEY / total number of sentences) -> Monetary values, including unit.
        """
        try:
            return SE.average_number_of_named_entities_money_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_money_ = \
                FoundationCollector.total_number_of_named_entities_money(SE)
            SE.average_number_of_named_entities_money_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_money_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_money_per_sentence_
    
    def average_number_of_named_entities_quantity_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are QUANTITY / total number of sentences) -> Measurements, as of weight or distance.
        """
        try:
            return SE.average_number_of_named_entities_quantity_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_quantity_ = \
                FoundationCollector.total_number_of_named_entities_quantity(SE)
            SE.average_number_of_named_entities_quantity_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_quantity_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_quantity_per_sentence_
    
    def average_number_of_named_entities_ordinal_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are ORDINAL / total number of sentences) -> “first”, “second”, etc.
        """
        try:
            return SE.average_number_of_named_entities_ordinal_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_ordinal_ = \
                FoundationCollector.total_number_of_named_entities_ordinal(SE)
            SE.average_number_of_named_entities_ordinal_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_ordinal_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_ordinal_per_sentence_
    
    def average_number_of_named_entities_cardinal_per_sentence(
        SE: object
        ) -> float:
        """
        returns the value of (total number of named entities that are CARDINAL / total number of sentences) -> Numerals that do not fall under another type.
        """
        try:
            return SE.average_number_of_named_entities_cardinal_per_sentence_ 
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_cardinal_ = \
                FoundationCollector.total_number_of_named_entities_cardinal(SE)
            SE.average_number_of_named_entities_cardinal_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_cardinal_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_cardinal_per_sentence_


class PartOfSpeech:
    """PartOfSpeech

    Parent class for features that are in the 'partofspeech' family.
    """
    
    def total_number_of_adjectives(
        SE: object
        ) -> float:
        """
        returns the number of adjectives
        """
        try: 
            return SE.total_number_of_adjectives_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_adjectives_ = pos_list.count("ADJ")
            return SE.total_number_of_adjectives_

    def total_number_of_unique_adjectives(
        SE: object
        ) -> float:
        """
        returns the number of unique adjectives
        """
        try: 
            return SE.total_number_of_unique_adjectives_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="ADJ"]
            SE.total_number_of_unique_adjectives_ = len([*set(pos_list)])
            return SE.total_number_of_unique_adjectives_

    def total_number_of_adpositions(
        SE: object
        ) -> float:
        """
        returns the number of adpositions
        """
        try: 
            return SE.total_number_of_adpositions_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_adpositions_ = pos_list.count("ADP")
            return SE.total_number_of_adpositions_
        
    def total_number_of_unique_adpositions(
        SE: object
        ) -> float:
        """
        returns the number of unique adpositions
        """
        try: 
            return SE.total_number_of_unique_adpositions_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="ADP"]
            SE.total_number_of_unique_adpositions_ = len([*set(pos_list)])
            return SE.total_number_of_unique_adpositions_

    def total_number_of_adverbs(
        SE: object
        ) -> float:
        """
        returns the number of adverbs
        """
        try: 
            return SE.total_number_of_adverbs_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_adverbs_ = pos_list.count("ADV")
            return SE.total_number_of_adverbs_
        
    def total_number_of_unique_adverbs(
        SE: object
        ) -> float:
        """
        returns the number of unique adverbs
        """
        try: 
            return SE.total_number_of_unique_adverbs_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="ADV"]
            SE.total_number_of_unique_adverbs_ = len([*set(pos_list)])
            return SE.total_number_of_unique_adverbs_

    def total_number_of_auxiliaries(
        SE: object
        ) -> float:
        """
        returns the number of auxiliaries
        """
        try: 
            return SE.total_number_of_auxiliaries_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_auxiliaries_ = pos_list.count("AUX")
            return SE.total_number_of_auxiliaries_
        
    def total_number_of_unique_auxiliaries(
        SE: object
        ) -> float:
        """
        returns the number of unique auxiliaries
        """
        try: 
            return SE.total_number_of_unique_auxiliaries_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="AUX"]
            SE.total_number_of_unique_auxiliaries_ = len([*set(pos_list)])
            return SE.total_number_of_unique_auxiliaries_

    def total_number_of_coordinating_conjunctions(
        SE: object
        ) -> float:
        """
        returns the number of coordinating_conjunctions
        """
        try: 
            return SE.total_number_of_coordinating_conjunctions_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_coordinating_conjunctions_ = pos_list.count("CCONJ")
            return SE.total_number_of_coordinating_conjunctions_

    def total_number_of_unique_coordinating_conjunctions(
        SE: object
        ) -> float:
        """
        returns the number of unique coordinating_conjunctions
        """
        try: 
            return SE.total_number_of_unique_coordinating_conjunctions_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="CCONJ"]
            SE.total_number_of_unique_coordinating_conjunctions_ = len([*set(pos_list)])
            return SE.total_number_of_unique_coordinating_conjunctions_

    def total_number_of_determiners(
        SE: object
        ) -> float:
        """
        returns the number of determiners
        """
        try: 
            return SE.total_number_of_determiners_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_determiners_ = pos_list.count("DET")
            return SE.total_number_of_determiners_

    def total_number_of_unique_determiners(
        SE: object
        ) -> float:
        """
        returns the number of unique determiners
        """
        try: 
            return SE.total_number_of_unique_determiners_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="DET"]
            SE.total_number_of_unique_determiners_ = len([*set(pos_list)])
            return SE.total_number_of_unique_determiners_

    def total_number_of_interjections(
        SE: object
        ) -> float:
        """
        returns the number of interjections
        """
        try: 
            return SE.total_number_of_interjections_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_interjections_ = pos_list.count("INTJ")
            return SE.total_number_of_interjections_

    def total_number_of_unique_interjections(
        SE: object
        ) -> float:
        """
        returns the number of unique interjections
        """
        try: 
            return SE.total_number_of_unique_interjections_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="INTJ"]
            SE.total_number_of_unique_interjections_ = len([*set(pos_list)])
            return SE.total_number_of_unique_interjections_

    def total_number_of_nouns(
        SE: object
        ) -> float:
        """
        returns the number of nouns
        """
        try: 
            return SE.total_number_of_nouns_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_nouns_ = pos_list.count("NOUN")
            return SE.total_number_of_nouns_

    def total_number_of_unique_nouns(
        SE: object
        ) -> float:
        """
        returns the number of unique nouns
        """
        try: 
            return SE.total_number_of_unique_nouns_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="NOUN"]
            SE.total_number_of_unique_nouns_ = len([*set(pos_list)])
            return SE.total_number_of_unique_nouns_

    def total_number_of_numerals(
        SE: object
        ) -> float:
        """
        returns the number of numerals
        """
        try: 
            return SE.total_number_of_numerals_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_numerals_ = pos_list.count("NUM")
            return SE.total_number_of_numerals_

    def total_number_of_unique_numerals(
        SE: object
        ) -> float:
        """
        returns the number of unique numerals
        """
        try: 
            return SE.total_number_of_unique_numerals_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="NUM"]
            SE.total_number_of_unique_numerals_ = len([*set(pos_list)])
            return SE.total_number_of_unique_numerals_

    def total_number_of_particles(
        SE: object
        ) -> float:
        """
        returns the number of particles
        """
        try: 
            return SE.total_number_of_particles_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_particles_ = pos_list.count("PART")
            return SE.total_number_of_particles_

    def total_number_of_unique_particles(
        SE: object
        ) -> float:
        """
        returns the number of unique particles
        """
        try: 
            return SE.total_number_of_unique_particles_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="PART"]
            SE.total_number_of_unique_particles_ = len([*set(pos_list)])
            return SE.total_number_of_unique_particles_

    def total_number_of_pronouns(
        SE: object
        ) -> float:
        """
        returns the number of pronouns
        """
        try: 
            return SE.total_number_of_pronouns_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_pronouns_ = pos_list.count("PRON")
            return SE.total_number_of_pronouns_

    def total_number_of_unique_pronouns(
        SE: object
        ) -> float:
        """
        returns the number of unique pronouns
        """
        try: 
            return SE.total_number_of_unique_pronouns_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="PRON"]
            SE.total_number_of_unique_pronouns_ = len([*set(pos_list)])
            return SE.total_number_of_unique_pronouns_

    def total_number_of_proper_nouns(
        SE: object
        ) -> float:
        """
        returns the number of proper nouns
        """
        try: 
            return SE.total_number_of_proper_nouns_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_proper_nouns_ = pos_list.count("PROPN")
            return SE.total_number_of_proper_nouns_

    def total_number_of_unique_proper_nouns(
        SE: object
        ) -> float:
        """
        returns the number of unique proper nouns
        """
        try: 
            return SE.total_number_of_unique_proper_nouns_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="PROPN"]
            SE.total_number_of_unique_proper_nouns_ = len([*set(pos_list)])
            return SE.total_number_of_unique_proper_nouns_

    def total_number_of_punctuations(
        SE: object
        ) -> float:
        """
        returns the number of punctuations
        """
        try: 
            return SE.total_number_of_punctuations_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_punctuations_ = pos_list.count("PUNCT")
            return SE.total_number_of_punctuations_

    def total_number_of_unique_punctuations(
        SE: object
        ) -> float:
        """
        returns the number of unique punctuations
        """
        try: 
            return SE.total_number_of_unique_punctuations_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="PUNCT"]
            SE.total_number_of_unique_punctuations_ = len([*set(pos_list)])
            return SE.total_number_of_unique_punctuations_

    def total_number_of_subordinating_conjunctions(
        SE: object
        ) -> float:
        """
        returns the number of subordinating conjunctions
        """
        try: 
            return SE.total_number_of_subordinating_conjunctions_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_subordinating_conjunctions_ = pos_list.count("SCONJ")
            return SE.total_number_of_subordinating_conjunctions_

    def total_number_of_unique_subordinating_conjunctions(
        SE: object
        ) -> float:
        """
        returns the number of unique subordinating conjunctions
        """
        try: 
            return SE.total_number_of_unique_subordinating_conjunctions_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="SCONJ"]
            SE.total_number_of_unique_subordinating_conjunctions_ = len([*set(pos_list)])
            return SE.total_number_of_unique_subordinating_conjunctions_

    def total_number_of_symbols(
        SE: object
        ) -> float:
        """
        returns the number of symbols
        """
        try: 
            return SE.total_number_of_symbols_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_symbols_ = pos_list.count("SYM")
            return SE.total_number_of_symbols_

    def total_number_of_unique_symbols(
        SE: object
        ) -> float:
        """
        returns the number of unique symbols
        """
        try: 
            return SE.total_number_of_unique_symbols_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="SYM"]
            SE.total_number_of_unique_symbols_ = len([*set(pos_list)])
            return SE.total_number_of_unique_symbols_

    def total_number_of_verbs(
        SE: object
        ) -> float:
        """
        returns the number of verbs
        """
        try: 
            return SE.total_number_of_verbs_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_verbs_ = pos_list.count("VERB")
            return SE.total_number_of_verbs_
    
    def total_number_of_unique_verbs(
        SE: object
        ) -> float:
        """
        returns the number of unique verbs
        """
        try: 
            return SE.total_number_of_unique_verbs_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="VERB"]
            SE.total_number_of_unique_verbs_ = len([*set(pos_list)])
            return SE.total_number_of_unique_verbs_

    def total_number_of_spaces(
        SE: object
        ) -> float:
        """
        returns the number of spaces
        """
        try: 
            return SE.total_number_of_spaces_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_spaces_ = pos_list.count("SPACE")
            return SE.total_number_of_spaces_

    def total_number_of_unique_spaces(
        SE: object
        ) -> float:
        """
        returns the number of unique spaces
        """
        try: 
            return SE.total_number_of_unique_spaces_
        except AttributeError:
            pos_list = [token.lemma_ for token in SE.doc if token.pos_=="SPACE"]
            SE.total_number_of_unique_spaces_ = len([*set(pos_list)])
            return SE.total_number_of_unique_spaces_



import math

from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class ReadFormula(FoundationCollector):
    """ReadFormula

    Parent class for features that are in the 'readformula' family.
    """

    def flesch_kincaid_reading_ease(
        SE: object,
        ) -> float:
        """
        returns reading difficulty value from flesch kincaid reading ease formula, where 100.00–90.00->5th grade (Very easy to read), 90.0–80.0->6th grade (Easy to read), 80.0–70.0->7th grade (Fairly easy to read), 70.0–60.0->8th & 9th grade (Plain English, 60.0–50.0->10th to 12th grade (Fairly difficult to read), 50.0–30.0->College (Difficult to read), 30.0–10.0->College graduate (Very difficult to read), 10.0–0.0->Professional (Extremely difficult to read)
        """
        try:
            return SE.flesch_kincaid_reading_ease_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_syllables_ = \
                FoundationCollector.total_number_of_syllables(SE)
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            average_number_of_syllables_per_word_ = \
                safe_division(
                    total_number_of_syllables_,
                    total_number_of_words_
                )
            # Calculate readability
            SE.flesch_kincaid_reading_ease_ = \
                206.835 - 1.015 * average_number_of_words_per_sentence_ - 84.6 * average_number_of_syllables_per_word_
            # Revert back to user options
            SE.options = original_options
            return SE.flesch_kincaid_reading_ease_
    
    def flesch_kincaid_grade_level(
        SE: object,
        ) -> float:
        """
        returns reading difficulty value that corresponds to US grade level
        """
        try:
            return SE.flesch_kincaid_grade_level_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_syllables_ = \
                FoundationCollector.total_number_of_syllables(SE)
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            average_number_of_syllables_per_word_ = \
                safe_division(
                    total_number_of_syllables_,
                    total_number_of_words_
                )
            # Calculate readability
            SE.flesch_kincaid_grade_level_ = \
                0.39 * average_number_of_words_per_sentence_ + 11.8 * average_number_of_syllables_per_word_ - 15.59
            # Revert back to user options
            SE.options = original_options
            return SE.flesch_kincaid_grade_level_
    
    def gunning_fog_index(
        SE: object,
        ) -> float:
        """
        returns gunning fog index, where 17->College graduate, 16->College senior, 15->College junior, 14->College sophomore, 13->College freshman, 12->High school senior, 11->High school junior, 10->High school sophomore, 9->High school freshman, 8->Eighth grade, 7->Seventh grade, 6->Sixth grade
        """
        try:
            return SE.gunning_fog_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_words_more_than_two_syllables_ = \
                FoundationCollector.total_number_of_words_more_than_two_syllables(SE)
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            average_number_of_words_more_than_two_syllables_per_word_ = \
                safe_division(
                    total_number_of_words_more_than_two_syllables_,
                    total_number_of_words_
                )
            # Calculate readability
            SE.gunning_fog_index_ = \
                0.4 * (average_number_of_words_per_sentence_ + 100 * average_number_of_words_more_than_two_syllables_per_word_)
            # Revert back to user options
            SE.options = original_options
            return SE.gunning_fog_index_
    
    def smog_index(
        SE: object,
        ) -> float:
        """
        returns reading difficulty value that corresponds to US grade level
        """
        try:
            return SE.smog_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_words_more_than_three_syllables_ = \
                FoundationCollector.total_number_of_words_more_than_three_syllables(SE)
            # Calculate readability
            SE.smog_index_ = \
                1.0430 * math.sqrt(total_number_of_words_more_than_three_syllables_ * safe_division(30, total_number_of_sentences_))
            # Revert back to user options
            SE.options = original_options
            return SE.smog_index_

    def coleman_liau_index(
        SE: object,
        ) -> float:
        """
        returns reading difficulty value that corresponds to US grade level
        """
        try:
            return SE.coleman_liau_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_characters_ = \
                FoundationCollector.total_number_of_characters(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            average_number_of_letters_per_100words_ = \
                safe_division(
                    total_number_of_characters_,
                    total_number_of_words_
                ) * 100
            average_number_of_sentences_per_100words_ = \
                safe_division(
                    total_number_of_sentences_,
                    total_number_of_words_
                ) * 100
            # Calculate readability
            SE.coleman_liau_index_ = \
                0.0588 * average_number_of_letters_per_100words_ - 0.296 * average_number_of_sentences_per_100words_ - 15.8
            # Revert back to user options
            SE.options = original_options
            return SE.coleman_liau_index_
    
    def automated_readability_index(
        SE: object,
        ) -> float:
        """
        returns reading difficulty value that corresponds to US grade level
        """
        try:
            return SE.automated_readability_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_characters_ = \
                FoundationCollector.total_number_of_characters(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            average_number_of_letters_per_word_ = \
                safe_division(
                    total_number_of_characters_,
                    total_number_of_words_
                )
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            # Calculate readability
            SE.automated_readability_index_ = \
                4.71 * average_number_of_letters_per_word_ + 0.5 * average_number_of_words_per_sentence_ - 21.43
            # Revert back to user options
            SE.options = original_options
            return SE.automated_readability_index_


from lftk.derivation.avgwordsent import AvgWordSent
from lftk.derivation.avgworddiff import AvgWordDiff
from lftk.derivation.avgentity   import AvgEntity
from lftk.derivation.typetokenratio import TypeTokenRatio
from lftk.derivation.avgpartofspeech import AvgPartOfSpeech
from lftk.derivation.lexicalvariation import LexicalVariation
from lftk.derivation.readformula import ReadFormula
from lftk.derivation.readtimeformula import ReadTimeFormula


class DerivationCollector(AvgWordSent, AvgWordDiff, AvgEntity, TypeTokenRatio, AvgPartOfSpeech, ReadFormula, ReadTimeFormula, LexicalVariation):
    pass 


from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class ReadTimeFormula(FoundationCollector):
    """ReadTimeFormula

    Parent class for features that are in the 'readtimeformula' family.
    """

    def reading_time_for_fast_readers(
        SE: object,
        ) -> float:
        """
        returns value of (total number of words / 300)
        """
        try:
            return SE.reading_time_for_fast_readers_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.reading_time_for_fast_readers_ = \
                safe_division(
                    total_number_of_words_,
                    300
                )
            # Revert back to user options
            SE.options = original_options
        return SE.reading_time_for_fast_readers_
    
    def reading_time_for_average_readers(
        SE: object,
        ) -> float:
        """
        returns value of (total number of words / 300)
        """
        try:
            return SE.reading_time_for_average_readers_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.reading_time_for_average_readers_ = \
                safe_division(
                    total_number_of_words_,
                    240
                )
            # Revert back to user options
            SE.options = original_options
        return SE.reading_time_for_average_readers_

    def reading_time_for_slow_readers(
        SE: object,
        ) -> float:
        """
        returns value of (total number of words / 300)
        """
        try:
            return SE.reading_time_for_slow_readers_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.reading_time_for_slow_readers_ = \
                safe_division(
                    total_number_of_words_,
                    175
                )
            # Revert back to user options
            SE.options = original_options
        return SE.reading_time_for_slow_readers_


import math

from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class LexicalVariation(FoundationCollector):
    """LexicalVariation

    Parent class for features that are in the 'lexicalvariation' family.
    """

    def simple_adjectives_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adjectives / total number of adjectives)
        """
        try:
            return SE.simple_adjectives_variation_
        except AttributeError:
            total_number_of_unique_adjectives_ = \
                FoundationCollector.total_number_of_unique_adjectives(SE)
            total_number_of_adjectives_ = \
                FoundationCollector.total_number_of_adjectives(SE)
            SE.simple_adjectives_variation_ = \
                safe_division(
                    total_number_of_unique_adjectives_,
                    total_number_of_adjectives_
                )
            return SE.simple_adjectives_variation_
    
    def root_adjectives_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adjectives / root(total number of adjectives))
        """
        try:
            return SE.root_adjectives_variation_
        except AttributeError:
            total_number_of_unique_adjectives_ = \
                FoundationCollector.total_number_of_unique_adjectives(SE)
            total_number_of_adjectives_ = \
                FoundationCollector.total_number_of_adjectives(SE)
            SE.root_adjectives_variation_ = \
                safe_division(
                    total_number_of_unique_adjectives_,
                    math.sqrt(total_number_of_adjectives_)
                )
            return SE.root_adjectives_variation_
        
    def corrected_adjectives_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adjectives / root(2*total number of adjectives))
        """
        try:
            return SE.corrected_adjectives_variation_
        except AttributeError:
            total_number_of_unique_adjectives_ = \
                FoundationCollector.total_number_of_unique_adjectives(SE)
            total_number_of_adjectives_ = \
                FoundationCollector.total_number_of_adjectives(SE)
            SE.corrected_adjectives_variation_ = \
                safe_division(
                    total_number_of_unique_adjectives_,
                    math.sqrt(2*total_number_of_adjectives_)
                )
            return SE.corrected_adjectives_variation_
    
    def simple_adpositions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adpositions / total number of adpositions)
        """
        try:
            return SE.simple_adpositions_variation_
        except AttributeError:
            total_number_of_unique_adpositions_ = \
                FoundationCollector.total_number_of_unique_adpositions(SE)
            total_number_of_adpositions_ = \
                FoundationCollector.total_number_of_adpositions(SE)
            SE.simple_adpositions_variation_ = \
                safe_division(
                    total_number_of_unique_adpositions_,
                    total_number_of_adpositions_
                )
            return SE.simple_adpositions_variation_
    
    def root_adpositions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adpositions / root(total number of adpositions))
        """
        try:
            return SE.root_adpositions_variation_
        except AttributeError:
            total_number_of_unique_adpositions_ = \
                FoundationCollector.total_number_of_unique_adpositions(SE)
            total_number_of_adpositions_ = \
                FoundationCollector.total_number_of_adpositions(SE)
            SE.root_adpositions_variation_ = \
                safe_division(
                    total_number_of_unique_adpositions_,
                    math.sqrt(total_number_of_adpositions_)
                )
            return SE.root_adpositions_variation_
        
    def corrected_adpositions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adpositions / root(2*total number of adpositions))
        """
        try:
            return SE.corrected_adpositions_variation_
        except AttributeError:
            total_number_of_unique_adpositions_ = \
                FoundationCollector.total_number_of_unique_adpositions(SE)
            total_number_of_adpositions_ = \
                FoundationCollector.total_number_of_adpositions(SE)
            SE.corrected_adpositions_variation_ = \
                safe_division(
                    total_number_of_unique_adpositions_,
                    math.sqrt(2*total_number_of_adpositions_)
                )
            return SE.corrected_adpositions_variation_
    
    def simple_adverbs_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adverbs / total number of adverbs)
        """
        try:
            return SE.simple_adverbs_variation_
        except AttributeError:
            total_number_of_unique_adverbs_ = \
                FoundationCollector.total_number_of_unique_adverbs(SE)
            total_number_of_adverbs_ = \
                FoundationCollector.total_number_of_adverbs(SE)
            SE.simple_adverbs_variation_ = \
                safe_division(
                    total_number_of_unique_adverbs_,
                    total_number_of_adverbs_
                )
            return SE.simple_adverbs_variation_
    
    def root_adverbs_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adverbs / root(total number of adverbs))
        """
        try:
            return SE.root_adverbs_variation_
        except AttributeError:
            total_number_of_unique_adverbs_ = \
                FoundationCollector.total_number_of_unique_adverbs(SE)
            total_number_of_adverbs_ = \
                FoundationCollector.total_number_of_adverbs(SE)
            SE.root_adverbs_variation_ = \
                safe_division(
                    total_number_of_unique_adverbs_,
                    math.sqrt(total_number_of_adverbs_)
                )
            return SE.root_adverbs_variation_
        
    def corrected_adverbs_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique adverbs / root(2*total number of adverbs))
        """
        try:
            return SE.corrected_adverbs_variation_
        except AttributeError:
            total_number_of_unique_adverbs_ = \
                FoundationCollector.total_number_of_unique_adverbs(SE)
            total_number_of_adverbs_ = \
                FoundationCollector.total_number_of_adverbs(SE)
            SE.corrected_adverbs_variation_ = \
                safe_division(
                    total_number_of_unique_adverbs_,
                    math.sqrt(2*total_number_of_adverbs_)
                )
            return SE.corrected_adverbs_variation_
    
    def simple_auxiliaries_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique auxiliaries / total number of auxiliaries)
        """
        try:
            return SE.simple_auxiliaries_variation_
        except AttributeError:
            total_number_of_unique_auxiliaries_ = \
                FoundationCollector.total_number_of_unique_auxiliaries(SE)
            total_number_of_auxiliaries_ = \
                FoundationCollector.total_number_of_auxiliaries(SE)
            SE.simple_auxiliaries_variation_ = \
                safe_division(
                    total_number_of_unique_auxiliaries_,
                    total_number_of_auxiliaries_
                )
            return SE.simple_auxiliaries_variation_
    
    def root_auxiliaries_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique auxiliaries / root(total number of auxiliaries))
        """
        try:
            return SE.root_auxiliaries_variation_
        except AttributeError:
            total_number_of_unique_auxiliaries_ = \
                FoundationCollector.total_number_of_unique_auxiliaries(SE)
            total_number_of_auxiliaries_ = \
                FoundationCollector.total_number_of_auxiliaries(SE)
            SE.root_auxiliaries_variation_ = \
                safe_division(
                    total_number_of_unique_auxiliaries_,
                    math.sqrt(total_number_of_auxiliaries_)
                )
            return SE.root_auxiliaries_variation_
        
    def corrected_auxiliaries_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique auxiliaries / root(2*total number of auxiliaries))
        """
        try:
            return SE.corrected_auxiliaries_variation_
        except AttributeError:
            total_number_of_unique_auxiliaries_ = \
                FoundationCollector.total_number_of_unique_auxiliaries(SE)
            total_number_of_auxiliaries_ = \
                FoundationCollector.total_number_of_auxiliaries(SE)
            SE.corrected_auxiliaries_variation_ = \
                safe_division(
                    total_number_of_unique_auxiliaries_,
                    math.sqrt(2*total_number_of_auxiliaries_)
                )
            return SE.corrected_auxiliaries_variation_
        
    def simple_coordinating_conjunctions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique coordinating conjunctions / total number of coordinating conjunctions)
        """
        try:
            return SE.simple_coordinating_conjunctions_variation_
        except AttributeError:
            total_number_of_unique_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_unique_coordinating_conjunctions(SE)
            total_number_of_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_coordinating_conjunctions(SE)
            SE.simple_coordinating_conjunctions_variation_ = \
                safe_division(
                    total_number_of_unique_coordinating_conjunctions_,
                    total_number_of_coordinating_conjunctions_
                )
            return SE.simple_coordinating_conjunctions_variation_
    
    def root_coordinating_conjunctions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique coordinating conjunctions / root(total number of coordinating conjunctions))
        """
        try:
            return SE.root_coordinating_conjunctions_variation_
        except AttributeError:
            total_number_of_unique_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_unique_coordinating_conjunctions(SE)
            total_number_of_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_coordinating_conjunctions(SE)
            SE.root_coordinating_conjunctions_variation_ = \
                safe_division(
                    total_number_of_unique_coordinating_conjunctions_,
                    math.sqrt(total_number_of_coordinating_conjunctions_)
                )
            return SE.root_coordinating_conjunctions_variation_
        
    def corrected_coordinating_conjunctions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique coordinating conjunctions / root(2*total number of coordinating conjunctions))
        """
        try:
            return SE.corrected_coordinating_conjunctions_variation_
        except AttributeError:
            total_number_of_unique_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_unique_coordinating_conjunctions(SE)
            total_number_of_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_coordinating_conjunctions(SE)
            SE.corrected_coordinating_conjunctions_variation_ = \
                safe_division(
                    total_number_of_unique_coordinating_conjunctions_,
                    math.sqrt(2*total_number_of_coordinating_conjunctions_)
                )
            return SE.corrected_coordinating_conjunctions_variation_
    
    def simple_determiners_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique determiners / total number of determiners)
        """
        try:
            return SE.simple_determiners_variation_
        except AttributeError:
            total_number_of_unique_determiners_ = \
                FoundationCollector.total_number_of_unique_determiners(SE)
            total_number_of_determiners_ = \
                FoundationCollector.total_number_of_determiners(SE)
            SE.simple_determiners_variation_ = \
                safe_division(
                    total_number_of_unique_determiners_,
                    total_number_of_determiners_
                )
            return SE.simple_determiners_variation_
    
    def root_determiners_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique determiners / root(total number of determiners))
        """
        try:
            return SE.root_determiners_variation_
        except AttributeError:
            total_number_of_unique_determiners_ = \
                FoundationCollector.total_number_of_unique_determiners(SE)
            total_number_of_determiners_ = \
                FoundationCollector.total_number_of_determiners(SE)
            SE.root_determiners_variation_ = \
                safe_division(
                    total_number_of_unique_determiners_,
                    math.sqrt(total_number_of_determiners_)
                )
            return SE.root_determiners_variation_
        
    def corrected_determiners_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique determiners / root(2*total number of determiners))
        """
        try:
            return SE.corrected_determiners_variation_
        except AttributeError:
            total_number_of_unique_determiners_ = \
                FoundationCollector.total_number_of_unique_determiners(SE)
            total_number_of_determiners_ = \
                FoundationCollector.total_number_of_determiners(SE)
            SE.corrected_determiners_variation_ = \
                safe_division(
                    total_number_of_unique_determiners_,
                    math.sqrt(2*total_number_of_determiners_)
                )
            return SE.corrected_determiners_variation_
    
    def simple_interjections_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique interjections / total number of interjections)
        """
        try:
            return SE.simple_interjections_variation_
        except AttributeError:
            total_number_of_unique_interjections_ = \
                FoundationCollector.total_number_of_unique_interjections(SE)
            total_number_of_interjections_ = \
                FoundationCollector.total_number_of_interjections(SE)
            SE.simple_interjections_variation_ = \
                safe_division(
                    total_number_of_unique_interjections_,
                    total_number_of_interjections_
                )
            return SE.simple_interjections_variation_
    
    def root_interjections_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique interjections / root(total number of interjections))
        """
        try:
            return SE.root_interjections_variation_
        except AttributeError:
            total_number_of_unique_interjections_ = \
                FoundationCollector.total_number_of_unique_interjections(SE)
            total_number_of_interjections_ = \
                FoundationCollector.total_number_of_interjections(SE)
            SE.root_interjections_variation_ = \
                safe_division(
                    total_number_of_unique_interjections_,
                    math.sqrt(total_number_of_interjections_)
                )
            return SE.root_interjections_variation_
        
    def corrected_interjections_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique interjections / root(2*total number of interjections))
        """
        try:
            return SE.corrected_interjections_variation_
        except AttributeError:
            total_number_of_unique_interjections_ = \
                FoundationCollector.total_number_of_unique_interjections(SE)
            total_number_of_interjections_ = \
                FoundationCollector.total_number_of_interjections(SE)
            SE.corrected_interjections_variation_ = \
                safe_division(
                    total_number_of_unique_interjections_,
                    math.sqrt(2*total_number_of_interjections_)
                )
            return SE.corrected_interjections_variation_
    
    def simple_nouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique nouns / total number of nouns)
        """
        try:
            return SE.simple_nouns_variation_
        except AttributeError:
            total_number_of_unique_nouns_ = \
                FoundationCollector.total_number_of_unique_nouns(SE)
            total_number_of_nouns_ = \
                FoundationCollector.total_number_of_nouns(SE)
            SE.simple_nouns_variation_ = \
                safe_division(
                    total_number_of_unique_nouns_,
                    total_number_of_nouns_
                )
            return SE.simple_nouns_variation_
    
    def root_nouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique nouns / root(total number of nouns))
        """
        try:
            return SE.root_nouns_variation_
        except AttributeError:
            total_number_of_unique_nouns_ = \
                FoundationCollector.total_number_of_unique_nouns(SE)
            total_number_of_nouns_ = \
                FoundationCollector.total_number_of_nouns(SE)
            SE.root_nouns_variation_ = \
                safe_division(
                    total_number_of_unique_nouns_,
                    math.sqrt(total_number_of_nouns_)
                )
            return SE.root_nouns_variation_
        
    def corrected_nouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique nouns / root(2*total number of nouns))
        """
        try:
            return SE.corrected_nouns_variation_
        except AttributeError:
            total_number_of_unique_nouns_ = \
                FoundationCollector.total_number_of_unique_nouns(SE)
            total_number_of_nouns_ = \
                FoundationCollector.total_number_of_nouns(SE)
            SE.corrected_nouns_variation_ = \
                safe_division(
                    total_number_of_unique_nouns_,
                    math.sqrt(2*total_number_of_nouns_)
                )
            return SE.corrected_nouns_variation_
    
    def simple_numerals_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique numerals / total number of numerals)
        """
        try:
            return SE.simple_numerals_variation_
        except AttributeError:
            total_number_of_unique_numerals_ = \
                FoundationCollector.total_number_of_unique_numerals(SE)
            total_number_of_numerals_ = \
                FoundationCollector.total_number_of_numerals(SE)
            SE.simple_numerals_variation_ = \
                safe_division(
                    total_number_of_unique_numerals_,
                    total_number_of_numerals_
                )
            return SE.simple_numerals_variation_
    
    def root_numerals_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique numerals / root(total number of numerals))
        """
        try:
            return SE.root_numerals_variation_
        except AttributeError:
            total_number_of_unique_numerals_ = \
                FoundationCollector.total_number_of_unique_numerals(SE)
            total_number_of_numerals_ = \
                FoundationCollector.total_number_of_numerals(SE)
            SE.root_numerals_variation_ = \
                safe_division(
                    total_number_of_unique_numerals_,
                    math.sqrt(total_number_of_numerals_)
                )
            return SE.root_numerals_variation_
        
    def corrected_numerals_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique numerals / root(2*total number of numerals))
        """
        try:
            return SE.corrected_numerals_variation_
        except AttributeError:
            total_number_of_unique_numerals_ = \
                FoundationCollector.total_number_of_unique_numerals(SE)
            total_number_of_numerals_ = \
                FoundationCollector.total_number_of_numerals(SE)
            SE.corrected_numerals_variation_ = \
                safe_division(
                    total_number_of_unique_numerals_,
                    math.sqrt(2*total_number_of_numerals_)
                )
            return SE.corrected_numerals_variation_
    
    def simple_particles_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique particles / total number of particles)
        """
        try:
            return SE.simple_particles_variation_
        except AttributeError:
            total_number_of_unique_particles_ = \
                FoundationCollector.total_number_of_unique_particles(SE)
            total_number_of_particles_ = \
                FoundationCollector.total_number_of_particles(SE)
            SE.simple_particles_variation_ = \
                safe_division(
                    total_number_of_unique_particles_,
                    total_number_of_particles_
                )
            return SE.simple_particles_variation_
    
    def root_particles_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique particles / root(total number of particles))
        """
        try:
            return SE.root_particles_variation_
        except AttributeError:
            total_number_of_unique_particles_ = \
                FoundationCollector.total_number_of_unique_particles(SE)
            total_number_of_particles_ = \
                FoundationCollector.total_number_of_particles(SE)
            SE.root_particles_variation_ = \
                safe_division(
                    total_number_of_unique_particles_,
                    math.sqrt(total_number_of_particles_)
                )
            return SE.root_particles_variation_
        
    def corrected_particles_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique particles / root(2*total number of particles))
        """
        try:
            return SE.corrected_particles_variation_
        except AttributeError:
            total_number_of_unique_particles_ = \
                FoundationCollector.total_number_of_unique_particles(SE)
            total_number_of_particles_ = \
                FoundationCollector.total_number_of_particles(SE)
            SE.corrected_particles_variation_ = \
                safe_division(
                    total_number_of_unique_particles_,
                    math.sqrt(2*total_number_of_particles_)
                )
            return SE.corrected_particles_variation_
    
    def simple_pronouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique pronouns / total number of pronouns)
        """
        try:
            return SE.simple_pronouns_variation_
        except AttributeError:
            total_number_of_unique_pronouns_ = \
                FoundationCollector.total_number_of_unique_pronouns(SE)
            total_number_of_pronouns_ = \
                FoundationCollector.total_number_of_pronouns(SE)
            SE.simple_pronouns_variation_ = \
                safe_division(
                    total_number_of_unique_pronouns_,
                    total_number_of_pronouns_
                )
            return SE.simple_pronouns_variation_
    
    def root_pronouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique pronouns / root(total number of pronouns))
        """
        try:
            return SE.root_pronouns_variation_
        except AttributeError:
            total_number_of_unique_pronouns_ = \
                FoundationCollector.total_number_of_unique_pronouns(SE)
            total_number_of_pronouns_ = \
                FoundationCollector.total_number_of_pronouns(SE)
            SE.root_pronouns_variation_ = \
                safe_division(
                    total_number_of_unique_pronouns_,
                    math.sqrt(total_number_of_pronouns_)
                )
            return SE.root_pronouns_variation_
        
    def corrected_pronouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique pronouns / root(2*total number of pronouns))
        """
        try:
            return SE.corrected_pronouns_variation_
        except AttributeError:
            total_number_of_unique_pronouns_ = \
                FoundationCollector.total_number_of_unique_pronouns(SE)
            total_number_of_pronouns_ = \
                FoundationCollector.total_number_of_pronouns(SE)
            SE.corrected_pronouns_variation_ = \
                safe_division(
                    total_number_of_unique_pronouns_,
                    math.sqrt(2*total_number_of_pronouns_)
                )
            return SE.corrected_pronouns_variation_
    
    def simple_proper_nouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique proper nouns / total number of proper nouns)
        """
        try:
            return SE.simple_proper_nouns_variation_
        except AttributeError:
            total_number_of_unique_proper_nouns_ = \
                FoundationCollector.total_number_of_unique_proper_nouns(SE)
            total_number_of_proper_nouns_ = \
                FoundationCollector.total_number_of_proper_nouns(SE)
            SE.simple_proper_nouns_variation_ = \
                safe_division(
                    total_number_of_unique_proper_nouns_,
                    total_number_of_proper_nouns_
                )
            return SE.simple_proper_nouns_variation_
    
    def root_proper_nouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique proper nouns / root(total number of proper nouns))
        """
        try:
            return SE.root_proper_nouns_variation_
        except AttributeError:
            total_number_of_unique_proper_nouns_ = \
                FoundationCollector.total_number_of_unique_proper_nouns(SE)
            total_number_of_proper_nouns_ = \
                FoundationCollector.total_number_of_proper_nouns(SE)
            SE.root_proper_nouns_variation_ = \
                safe_division(
                    total_number_of_unique_proper_nouns_,
                    math.sqrt(total_number_of_proper_nouns_)
                )
            return SE.root_proper_nouns_variation_
        
    def corrected_proper_nouns_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique proper nouns / root(2*total number of proper nouns))
        """
        try:
            return SE.corrected_proper_nouns_variation_
        except AttributeError:
            total_number_of_unique_proper_nouns_ = \
                FoundationCollector.total_number_of_unique_proper_nouns(SE)
            total_number_of_proper_nouns_ = \
                FoundationCollector.total_number_of_proper_nouns(SE)
            SE.corrected_proper_nouns_variation_ = \
                safe_division(
                    total_number_of_unique_proper_nouns_,
                    math.sqrt(2*total_number_of_proper_nouns_)
                )
            return SE.corrected_proper_nouns_variation_
    
    def simple_punctuations_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique punctuations / total number of punctuations)
        """
        try:
            return SE.simple_punctuations_variation_
        except AttributeError:
            total_number_of_unique_punctuations_ = \
                FoundationCollector.total_number_of_unique_punctuations(SE)
            total_number_of_punctuations_ = \
                FoundationCollector.total_number_of_punctuations(SE)
            SE.simple_punctuations_variation_ = \
                safe_division(
                    total_number_of_unique_punctuations_,
                    total_number_of_punctuations_
                )
            return SE.simple_punctuations_variation_
    
    def root_punctuations_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique punctuations / root(total number of punctuations))
        """
        try:
            return SE.root_punctuations_variation_
        except AttributeError:
            total_number_of_unique_punctuations_ = \
                FoundationCollector.total_number_of_unique_punctuations(SE)
            total_number_of_punctuations_ = \
                FoundationCollector.total_number_of_punctuations(SE)
            SE.root_punctuations_variation_ = \
                safe_division(
                    total_number_of_unique_punctuations_,
                    math.sqrt(total_number_of_punctuations_)
                )
            return SE.root_punctuations_variation_
        
    def corrected_punctuations_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique punctuations / root(2*total number of punctuations))
        """
        try:
            return SE.corrected_punctuations_variation_
        except AttributeError:
            total_number_of_unique_punctuations_ = \
                FoundationCollector.total_number_of_unique_punctuations(SE)
            total_number_of_punctuations_ = \
                FoundationCollector.total_number_of_punctuations(SE)
            SE.corrected_punctuations_variation_ = \
                safe_division(
                    total_number_of_unique_punctuations_,
                    math.sqrt(2*total_number_of_punctuations_)
                )
            return SE.corrected_punctuations_variation_
    
    def simple_subordinating_conjunctions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique subordinating conjunctions / total number of subordinating conjunctions)
        """
        try:
            return SE.simple_subordinating_conjunctions_variation_
        except AttributeError:
            total_number_of_unique_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_unique_subordinating_conjunctions(SE)
            total_number_of_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_subordinating_conjunctions(SE)
            SE.simple_subordinating_conjunctions_variation_ = \
                safe_division(
                    total_number_of_unique_subordinating_conjunctions_,
                    total_number_of_subordinating_conjunctions_
                )
            return SE.simple_subordinating_conjunctions_variation_
    
    def root_subordinating_conjunctions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique subordinating conjunctions / root(total number of subordinating conjunctions))
        """
        try:
            return SE.root_subordinating_conjunctions_variation_
        except AttributeError:
            total_number_of_unique_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_unique_subordinating_conjunctions(SE)
            total_number_of_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_subordinating_conjunctions(SE)
            SE.root_subordinating_conjunctions_variation_ = \
                safe_division(
                    total_number_of_unique_subordinating_conjunctions_,
                    math.sqrt(total_number_of_subordinating_conjunctions_)
                )
            return SE.root_subordinating_conjunctions_variation_
        
    def corrected_subordinating_conjunctions_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique subordinating conjunctions / root(2*total number of subordinating conjunctions))
        """
        try:
            return SE.corrected_subordinating_conjunctions_variation_
        except AttributeError:
            total_number_of_unique_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_unique_subordinating_conjunctions(SE)
            total_number_of_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_subordinating_conjunctions(SE)
            SE.corrected_subordinating_conjunctions_variation_ = \
                safe_division(
                    total_number_of_unique_subordinating_conjunctions_,
                    math.sqrt(2*total_number_of_subordinating_conjunctions_)
                )
            return SE.corrected_subordinating_conjunctions_variation_
    
    def simple_symbols_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique symbols / total number of symbols)
        """
        try:
            return SE.simple_symbols_variation_
        except AttributeError:
            total_number_of_unique_symbols_ = \
                FoundationCollector.total_number_of_unique_symbols(SE)
            total_number_of_symbols_ = \
                FoundationCollector.total_number_of_symbols(SE)
            SE.simple_symbols_variation_ = \
                safe_division(
                    total_number_of_unique_symbols_,
                    total_number_of_symbols_
                )
            return SE.simple_symbols_variation_
    
    def root_symbols_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique symbols / root(total number of symbols))
        """
        try:
            return SE.root_symbols_variation_
        except AttributeError:
            total_number_of_unique_symbols_ = \
                FoundationCollector.total_number_of_unique_symbols(SE)
            total_number_of_symbols_ = \
                FoundationCollector.total_number_of_symbols(SE)
            SE.root_symbols_variation_ = \
                safe_division(
                    total_number_of_unique_symbols_,
                    math.sqrt(total_number_of_symbols_)
                )
            return SE.root_symbols_variation_
        
    def corrected_symbols_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique symbols / root(2*total number of symbols))
        """
        try:
            return SE.corrected_symbols_variation_
        except AttributeError:
            total_number_of_unique_symbols_ = \
                FoundationCollector.total_number_of_unique_symbols(SE)
            total_number_of_symbols_ = \
                FoundationCollector.total_number_of_symbols(SE)
            SE.corrected_symbols_variation_ = \
                safe_division(
                    total_number_of_unique_symbols_,
                    math.sqrt(2*total_number_of_symbols_)
                )
            return SE.corrected_symbols_variation_
    
    def simple_verbs_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique verbs / total number of verbs)
        """
        try:
            return SE.simple_verbs_variation_
        except AttributeError:
            total_number_of_unique_verbs_ = \
                FoundationCollector.total_number_of_unique_verbs(SE)
            total_number_of_verbs_ = \
                FoundationCollector.total_number_of_verbs(SE)
            SE.simple_verbs_variation_ = \
                safe_division(
                    total_number_of_unique_verbs_,
                    total_number_of_verbs_
                )
            return SE.simple_verbs_variation_
    
    def root_verbs_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique verbs / root(total number of verbs))
        """
        try:
            return SE.root_verbs_variation_
        except AttributeError:
            total_number_of_unique_verbs_ = \
                FoundationCollector.total_number_of_unique_verbs(SE)
            total_number_of_verbs_ = \
                FoundationCollector.total_number_of_verbs(SE)
            SE.root_verbs_variation_ = \
                safe_division(
                    total_number_of_unique_verbs_,
                    math.sqrt(total_number_of_verbs_)
                )
            return SE.root_verbs_variation_
        
    def corrected_verbs_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique verbs / root(2*total number of verbs))
        """
        try:
            return SE.corrected_verbs_variation_
        except AttributeError:
            total_number_of_unique_verbs_ = \
                FoundationCollector.total_number_of_unique_verbs(SE)
            total_number_of_verbs_ = \
                FoundationCollector.total_number_of_verbs(SE)
            SE.corrected_verbs_variation_ = \
                safe_division(
                    total_number_of_unique_verbs_,
                    math.sqrt(2*total_number_of_verbs_)
                )
            return SE.corrected_verbs_variation_
        
    def simple_spaces_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique spaces / total number of spaces)
        """
        try:
            return SE.simple_spaces_variation_
        except AttributeError:
            total_number_of_unique_spaces_ = \
                FoundationCollector.total_number_of_unique_spaces(SE)
            total_number_of_spaces_ = \
                FoundationCollector.total_number_of_spaces(SE)
            SE.simple_spaces_variation_ = \
                safe_division(
                    total_number_of_unique_spaces_,
                    total_number_of_spaces_
                )
            return SE.simple_spaces_variation_
    
    def root_spaces_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique spaces / root(total number of spaces))
        """
        try:
            return SE.root_spaces_variation_
        except AttributeError:
            total_number_of_unique_spaces_ = \
                FoundationCollector.total_number_of_unique_spaces(SE)
            total_number_of_spaces_ = \
                FoundationCollector.total_number_of_spaces(SE)
            SE.root_spaces_variation_ = \
                safe_division(
                    total_number_of_unique_spaces_,
                    math.sqrt(total_number_of_spaces_)
                )
            return SE.root_spaces_variation_
        
    def corrected_spaces_variation(
        SE: object,
        ) -> float:
        """
        returns value of (total number of unique spaces / root(2*total number of spaces))
        """
        try:
            return SE.corrected_spaces_variation_
        except AttributeError:
            total_number_of_unique_spaces_ = \
                FoundationCollector.total_number_of_unique_spaces(SE)
            total_number_of_spaces_ = \
                FoundationCollector.total_number_of_spaces(SE)
            SE.corrected_spaces_variation_ = \
                safe_division(
                    total_number_of_unique_spaces_,
                    math.sqrt(2*total_number_of_spaces_)
                )
            return SE.corrected_spaces_variation_